/*
  mm_profile.c
  Kernel module for 2.2.x Kernels
  
  This file is part of mergemem by Philipp Richter & Philipp Reisner

  mergemem is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  mergemem is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with mergemem; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifdef HAVE_AUTOCONF
#include <linux/autoconf.h> /* for CONFIG_MODVERSIONS */
#endif  
#include <asm/smp.h>
#include <asm/current.h>
#include <linux/config.h>
#include <linux/module.h>
#ifdef CONFIG_MODVERSIONS
#include <linux/modversions.h>
#endif  
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/swap.h>
#include <linux/miscdevice.h>

#include <asm/page.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/pgtable.h>

#define MODULE_NAME "mm_profile: "

EXPORT_NO_SYMBOLS;
MODULE_AUTHOR("Philipp Reisner <e9525415@stud2.tuwien.ac.at>");
MODULE_DESCRIPTION("mm_profile - profiles mergemem");

extern void (*mergemem_profile_hook1)(mem_map_t * map);
extern void (*mergemem_profile_hook2)(mem_map_t * map);
unsigned long mergemem_stat_saved = 0L;

/* proc file ops */
static int mmprof_proc_open(struct inode*, struct file*);
static int mmprof_proc_close(struct inode*,struct file*);
static ssize_t mmprof_proc_read(struct file*, char*, size_t, loff_t* );

static struct file_operations mmprof_proc_fops = 
{
  NULL,                   /* lseek - default */
  mmprof_proc_read,     /* read */
  NULL,                   /* write  */
  NULL,                   /* readdir */
  NULL,                   /* select - default */
  NULL,                   /* ioctl - default */
  NULL,                   /* mmap */
  mmprof_proc_open,     /* open code */
  NULL,                   /* flush */
  mmprof_proc_close,    /* release code */
  NULL                    /* can't fsync */
};

static struct inode_operations mmprof_proc_iops =
{
  &mmprof_proc_fops,    /* my own file-ops */
  NULL,                   /* create */
  NULL,                   /* lookup */
  NULL,                   /* link */
  NULL,                   /* unlink */
  NULL,                   /* symlink */
  NULL,                   /* mkdir */
  NULL,                   /* rmdir */
  NULL,                   /* mknod */
  NULL,                   /* rename */
  NULL,                   /* readlink */
  NULL,                   /* follow_link */
  NULL,                   /* readpage */
  NULL,                   /* writepage */
  NULL,                   /* bmap */
  NULL,                   /* truncate */
  NULL                    /* permission */
};
  
struct proc_dir_entry mmprof_proc_dir = 
{
  0, 10, "mm_profile",
  S_IFREG | S_IRUGO , 1, 0, 0,
  0, &mmprof_proc_iops,
  NULL, NULL,
  NULL,
  NULL, NULL
};

/* proc file operations */
static int mmprof_proc_open(struct inode *inode, struct file *file)
{
  MOD_INC_USE_COUNT;
  return 0;
}

static int mmprof_proc_close(struct inode *inode, struct file *file)
{
  MOD_DEC_USE_COUNT;
  return 0;
}

static ssize_t mmprof_proc_read(struct file *file, char *buf, size_t count, loff_t *unused)
{
  int fpos = file->f_pos;
  char buffer[200];

  sprintf(buffer,
          "version       : %d\n"
	  "saved         : %lu pages, %lu kb\n"
	  , MOD_VERSION,
	  mergemem_stat_saved, mergemem_stat_saved * PAGE_SIZE / 1024);

  if(fpos > strlen(buffer) - 1) return 0;
  if((strlen(buffer) - fpos) < count) 
    count = strlen(buffer) - fpos;
  copy_to_user(buf, buffer+fpos, count);
  file->f_pos = file->f_pos+count;

  return count;
}

/* Additional Bit for the page flags 11...30 are unused */

#define PG_mergemem_saved    28

void profiler_1(mem_map_t * map)
{
  if (test_bit(PG_mergemem_saved,&map->flags)) 
    {
      map->unused++;
    }
  else
    {
      map->flags |= (1 << PG_mergemem_saved); 
      map->unused=1;
    }
  mergemem_stat_saved++;
}

void profiler_2(mem_map_t * map)   
{
  if (test_bit(PG_mergemem_saved,&map->flags)) 
    {
      map->unused--;
      if(map->unused == 0) map->flags &= ~(1 << PG_mergemem_saved);
      mergemem_stat_saved--;
    }
}

int init_module()
{
  /* register proc */
  if(proc_register(&proc_root, &mmprof_proc_dir))
  {
    printk(MODULE_NAME"unable to register proc file.\n");
    return -EIO;
  }

  mergemem_profile_hook1 = profiler_1;
  mergemem_profile_hook2 = profiler_2;
  mergemem_stat_saved=0L;

  printk(MODULE_NAME"module initialised. Version: %d\n",MOD_VERSION);
  return 0;
}

void cleanup_module()
{
  mergemem_profile_hook1 = NULL;
  mergemem_profile_hook2 = NULL;
  proc_unregister(&proc_root, mmprof_proc_dir.low_ino);
  printk(MODULE_NAME"module released.\n");
}
